-- microYASEP.xx.vhdl
-- créé le 19 février 2012 par Yann Guidon / whygee@f-cpu.org

-- 01 : resynchronisation du reset local  et changement de phase
-- 02 : PC, flags, registres de pipeline & instructions
-- 03 : mise en place des ports et de la RAM instructions
-- 04 : mise en place du banc de registres
--        souci : calculer PC+1 et PC+2 ?
--        solution : fournir juste PC, et NPC en bout
--          de pipeline pour les instructions CALL/CALL2/NPC
-- 05 : PC résolu, CALL possible, ajout chemin d'adresse et de données
-- 06 : ALU/ROP2
-- 07 : manque le chemin de contrôle, conditions, WB_en

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
library work;
    use work.yasep_definitions;

entity microYASEP is
port(

 -- La mémoire instruction est en blockRAM, hors du "coeur".
 -- elle est lue en continu, la RAM active en permanence
 -- donc pas de signal RAM_read_en <= '0';
  signal RAM_out: in SLV16; -- port de lecture de la RAM instruction
  signal NPC: inout std_logic_vector(YASEP_SIZE-1 downto 1);   -- nouvelle valeur du compteur d'instructions   /!\ LSB=0

  clk, ext_reset : in std_logic);
end microYASEP;

architecture behav of microYASEP is

  signal phase, int_reset, condition, condition2, WB_en, Carry: std_logic;
  signal RAM_addr : std_logic_vector(addr_range);
  signal opcode: SLV6;
  signal cond_func, form: SLV2;
  signal group: SLV3;
  signal Field_SI4, Field_SND, DST: SLV4;
  signal SI4, RSI4_out, Reg_SI4,
         SND, RSND_out,
         ActualA, ActualB, ROP2_RS1, ROP2_RS2,
         ASU, ROP2, Result: SLVY;
  signal sumAux : unsigned(YASEP_SIZE+1 downto 0);
  signal PC, PC1 : std_logic_vector(YASEP_SIZE-1 downto 1);

  -- Les registres :
  signal  R1, R2, R3, R4, R5, 
      A1, D1, A2, D2, A3, D3, A4, D4, A5, D5 : SLVY;

begin

------------------------- Zone asynchrone -------------------------

-- Lecture des registres :

  RSND_out <= PC & '0'  when RAM_out(11 downto 8)=REG_PC
              else R1   when RAM_out(11 downto 8)=REG_R1
              else R2   when RAM_out(11 downto 8)=REG_R2
              else R3   when RAM_out(11 downto 8)=REG_R3
              else R4   when RAM_out(11 downto 8)=REG_R4
              else R5   when RAM_out(11 downto 8)=REG_R5
              else D1   when RAM_out(11 downto 8)=REG_D1
              else A1   when RAM_out(11 downto 8)=REG_A1
              else D2   when RAM_out(11 downto 8)=REG_D2
              else A2   when RAM_out(11 downto 8)=REG_A2
              else D3   when RAM_out(11 downto 8)=REG_D3
              else A3   when RAM_out(11 downto 8)=REG_A3
              else D4   when RAM_out(11 downto 8)=REG_D4
              else A4   when RAM_out(11 downto 8)=REG_A4
              else D5   when RAM_out(11 downto 8)=REG_D5
              else A5;--when RAM_out(11 downto 8)=REG_A5

  RSI4_out <= PC & '0'  when RAM_out(15 downto 12)=REG_PC
              else R1   when RAM_out(15 downto 12)=REG_R1
              else R2   when RAM_out(15 downto 12)=REG_R2
              else R3   when RAM_out(15 downto 12)=REG_R3
              else R4   when RAM_out(15 downto 12)=REG_R4
              else R5   when RAM_out(15 downto 12)=REG_R5
              else D1   when RAM_out(15 downto 12)=REG_D1
              else A1   when RAM_out(15 downto 12)=REG_A1
              else D2   when RAM_out(15 downto 12)=REG_D2
              else A2   when RAM_out(15 downto 12)=REG_A2
              else D3   when RAM_out(15 downto 12)=REG_D3
              else A3   when RAM_out(15 downto 12)=REG_A3
              else D4   when RAM_out(15 downto 12)=REG_D4
              else A4   when RAM_out(15 downto 12)=REG_A4
              else D5   when RAM_out(15 downto 12)=REG_D5
              else A5;--when RAM_out(15 downto 12)=REG_A5

-- chemin de contrôle : calcule WB_en

WB_en <= '1' when FlagValidInstruction(to_integer(opcode))='1'
              and condition='1'
    else '0';

condition <= FlagNoCondition(to_integer(opcode))
               when ((RAM_out(5) xor Condition2)='0'   -- condition is false
                 and  form = "11") -- extended / conditional instruction
   else '1'; -- true otherwise

Condition2 <= R1(to_integer(FieldSI4))  when cond_func="01"
         else  '1'                      when cond_func="00" and Field_SI4=REG_PC
         else Carry                     when cond_func="10" and Field_SI4=REG_PC
         else SRC4Zero                  when cond_func="00"
         else SRC4(0)                   when cond_func="10"
         else SRC4(15)             --   when cond_func="11"  -- toutes les combinaisons ne sont pas couvertes

cond_func <= RAM_out(7 downto 6);


-- chemin d'adresse :
  DST <= Field_SND when form(0) = '0'
    else Field_SI4 when form(1) = '0'
    else RAM_out(R_SND); -- correspond à DST3 car c'est le 2è mot

-- mise à jour du compteur de programme
  PC1 <= std_logic_vector(unsigned(PC)+1);
  NPC <= SI4(YASEP_SIZE-1 downto 1) when phase='1' and WB_en='1' and DST=REG_PC and opcode=Op_CALL
    else ASU(YASEP_SIZE-1 downto 1) when phase='1' and WB_en='1' and DST=REG_PC and opcode=Op_CALL2
    else PC1;

-- /!\ bien faire attention à l'échange qui se produit lors de CALL/CALL2,
--         entre NPC et Result. (ça se passe avec PC1 et SI4/ASU

  group <= opcode(2 downto 0);
  Result <= ASU         when group=GROUP_ASU
       else ROP2        when group=GROUP_ROP2
       else PC1         when opcode=Op_CALL or opcode=Op_CALL2
       else GetReadPort when opcode=Op_GET
       else SI4;  -- toutes les autres instructions, comme MOV

-- Chemin de données : sélection de SI4
  SI4 <=  RAM_out when form = "01" -- long immédiat
     else std_logic_vector(YASEP_SIZE-1 downto 4 => FieldSI4(3))   -- extension de signe
         & std_logic_vector(FieldSI4) when
               (form = "10")                     -- short
            or (form = "11" and  RAM_out(1)='1') -- extended
     else RSI4_out;

----------------- ALU/ROP2 -----------------

--         ROP2_S1  ROP2_S2  ROP2_X2  Addsub  compare_signed
-- ADD        X        X        X       0           0
-- SUB        X        X        X       1           0
-- compare                                (== SUB, only carry-out used)
--    signed  X        X        X       1           1
--  unsigned  X        X        X       1           0
-- AND        1        0        0       0           0
-- ANDN       1        0        0       1           0
-- NAND       1        0        1       0           0
-- OR         X        1        0       0           0
-- ORN        X        1        0       1           0
-- NOR        X        1        1       0           0
-- XOR        0        0        0       0           0
-- XORN       0        0        0       1           0

ROP2_S1 <= '1' when
    opcode=Op_AND  or
    opcode=Op_ANDN or
    opcode=Op_NAND
  else '0';

ROP2_S2 <= '1' when
    opcode=Op_OR  or
    opcode=Op_ORN or
    opcode=Op_NOR
  else '0';

ROP2_X2 <= '1' when
    opcode=Op_XOR  or
    opcode=Op_XORN
  else '0';

Addsub <= '1' when
    opcode=Op_SUB or
    opcode=Op_CMPS or
    opcode=Op_CMPU or
    opcode=Op_ANDN or
    opcode=Op_ORN or
    opcode=Op_XORN
  else '0';

compare_signed <='1' when
    opcode=Op_CMPS or
    opcode=Op_SMAX
  else '0';

-- XOR préliminaire et ajustement du signe :
ActualB <= ( SND(YASEP_SIZE-1) xor (Addsub and (not compare_signed)) )
         & ( SND(YASEP_SIZE-2 downto 0) xor (YASEP_SIZE-2 downto 0 =>Addsub);

   -- XOR the MSB here too for the signed comparisons :
ActualA <= ( SI4(YASEP_SIZE-1) xor compare_signed )
         &   SI4(YASEP_SIZE-2 downto 0);

-- Add/Sub
sumAux <= unsigned('0' & ActualA & '1') + unsigned('0' & ActualB & Addsub);

Carry_out <= std_logic(sumAux(YASEP_SIZE+1));
ASU <= std_logic_vector(sumAux(YASEP_SIZE downto 1));

-- ROP2 :
-- First MUX
ROP2_RS1 <= ActualA and ActualB when ROP2_S1='1'
    else    ActualA xor ActualB;

ROP2_RS2 <= ActualA  or ActualB when ROP2_S2='1'
       else ROP2_RS1;

-- last XOR
ROP2 <= ROP2_RS2 xor (ROP2_OUT'range => ROP2_X2);


------------------------- Zone synchrone -------------------------

  process (clk) begin
    -- resynchronisation du reset local
    if rising_edge(clk) then
      int_reset <= ext_reset;
    end if;
  end process;


  process (clk, int_reset) begin
    if int_reset='0' then
      phase <='1'; -- on commence en phase "execute" pour mettre PC à 0
      -- met l'instruction "MOV 0 PC" dans les registres concernés :
      opcode <= Op_MOV;
      form   <= FORM_RI4;
      Field_SND <= REG_PC;
      Field_SI4 <= "0000";
    else

      if rising_edge(clk) then
        -- changement de phase
        if phase='0' then
-- phase "fetch" : on écrit la sortie de la RAM dans des registres temporaires
          phase <= '1';
              form  <= RAM_out(1 downto 0);
             opcode <= RAM_out(R_OPCODE);
          Field_SND <= RAM_out(R_SND);
              SND   <= RSND_out;

          -- gère la longueur de l'instruction
          if RAM_out(0)='1' then
            -- instruction longue : Lire la suite
            PC <= NPC;

            if form=FORM_EXT then -- écriture spéculative, on ne sait pas quel champ sera utilisé
              Reg_SI4 <= RSI4_out;  -- champ inutilité en forme IMM16
            end if;
            Field_SI4 <= RAM_out(R_SI4);

          else
            -- instruction courte : Faire une pause d'une phase
            if form=FORM_RI4 then -- ne latche que le nécessaire
              Field_SI4 <= RAM_out(R_SI4);
            else
              Reg_SI4 <= RSI4_out;
            end if;
          end if;

        else
-- phase "execute" : on calcule et latche des valeurs, adresses et conditions
          phase <= '0';

          if WB_en='1' then
            if FlagChangeCarry(to_integer(opcode))='1' then
              Carry <= Carry_out;
            end if;
            if FlagNoWriteBack(to_integer(opcode))='0' then
              case DST is
                when REG_R1 => R1 <= Result;
                when REG_R2 => R2 <= Result;
                when REG_R3 => R3 <= Result;
                when REG_R4 => R4 <= Result;
                when REG_R5 => R5 <= Result;
                when REG_A1 => A1 <= Result;
                when REG_D1 => D1 <= Result;
                when REG_A2 => A2 <= Result;
                when REG_D2 => D2 <= Result;
                when REG_A3 => A3 <= Result;
                when REG_D3 => D3 <= Result;
                when REG_A4 => A4 <= Result;
                when REG_D4 => D4 <= Result;
                when REG_A5 => A5 <= Result;
                when REG_D5 => D5 <= Result;
                when others => ; rien;
              end case;
            end if;
          end if;
          PC <= NPC;  -- RAM_addr mis à jour en même temps
        end if;
      end if;
    end if;
  end process;

end behav;
